from urllib.parse import urlencode
from collections import namedtuple

from tornado.escape import json_decode

from talkzoho.service_client import ServiceClient
from talkzoho.crm.crm_resource import CRMResource
from talkzoho import logger


ModuleMap = namedtuple('ModuleMap', 'canonical_name, singular_alias, plural_alias')  # noqa


class CRMClient(ServiceClient):

    SCOPE              = 'crmapi'
    MAX_PAGE_SIZE      = 200

    def __init__(self, *args, use_module_aliases=False, **kwargs):
        super().__init__(*args, **kwargs)
        self.use_module_aliases = use_module_aliases

    @property
    def base_url(self):
        return 'https://crm.zoho.{region}/crm/private/json'.format(
            region=self.region)

    @property
    def base_query(self):
        return {
            'scope': self.SCOPE,
            'authtoken': self.auth_token}

    async def get_module_maps(self):
        # TODO: only do if map alias is True
        url  = '{base_url}/Info/getModules?{query}'.format(
            base_url=self.base_url,
            query=urlencode(self.base_query))

        logger.info('GET: {}'.format(url))
        response = await self.http_client.fetch(url)

        body = json_decode(response.body.decode('utf-8'))
        maps = body['response']['result']['row']

        return [ModuleMap(
            canonical_name=m['content'],
            singular_alias=m['sl'],
            plural_alias=m['pl']) for m in maps]

    def __getattr__(self, attr):
        """
        Translates attribute access to Zoho CRM module names.
        e.g. leads becomes Leads and custom_module_8 becomes CustomModule8
        """
        components  = attr.split('_')
        module_name = ''.join(c.title() for c in components)
        return CRMResource(service=self, name=module_name)



from typing import Union, Optional

from talkzoho.projects.base_resource import BaseResource


class PortalResource(BaseResource):

    def module_url(self, module_name):
        return '{base_url}/portal/{portal_id}/{module}/'.format(
            base_url=self.service.base_url,
            portal_id=self.portal_id,
            module=module_name)

    async def get(self,
                  id: Union[int, str], *,
                  portal_id: Union[str, int],
                  columns=None):
        self.portal_id = portal_id
        return await super().get(id=id, columns=columns)

    async def filter(self, *,
                     portal_id: Union[str, int],
                     term: Optional[str]=None,
                     columns: Optional[list]=None,
                     offset: int=0,
                     limit: Optional[int]=None):
        self.portal_id = portal_id
        return await super().filter(
            term=term,
            columns=columns,
            offset=offset,
            limit=limit)

    async def insert(self, record: dict):
        self.portal_id = record.pop('portal_id')
        return await super().insert(record=record)

    async def update(self, record: dict):
        self.portal_id = record.pop('portal_id')
        return await super().update(record=record)

    async def delete(self, id: Union[int, str], *, portal_id: Union[str, int]):
        self.portal_id = portal_id
        return await super().delete(id=id)



from typing import Union, Optional

from tornado.web import HTTPError

from talkzoho.projects.base_resource import BaseResource


class LogResource(BaseResource):

    def module_url(self, module_name):
        if self.component_type is None or self.component_type == 'general':
            return '{base_url}/portal/{portal_id}/projects/{project_id}/{module}/'.format(  # noqa
                base_url=self.service.base_url,
                portal_id=self.portal_id,
                project_id=self.project_id,
                module=module_name)
        else:
            return '{base_url}/portal/{portal_id}/projects/{project_id}/{component_type}/{component_id}/{module}/'.format(  # noqa
                base_url=self.service.base_url,
                portal_id=self.portal_id,
                project_id=self.project_id,
                component_type=self.component_type,
                component_id=self.component_id,
                module=module_name)

    async def get(self,
                  id: Union[int, str], *,
                  portal_id: Union[str, int],
                  project_id: Union[str, int],
                  component_type: Optional[str]=None,
                  columns=None):
        self.portal_id      = portal_id
        self.project_id     = project_id
        self.component_type = None

        log = await super().get(id=id, columns=columns)

        if component_type and log['component_type'] == component_type:
            raise HTTPError(404, 'No log with type {} found with that id. '
                                 'A {} log with that id does exist though.'.format(component_type, log['component_type']))  # noqa: E501

        return log

    async def filter(self, *,
                     portal_id: Union[str, int],
                     project_id: Union[str, int],
                     term: Optional[str]=None,
                     columns: Optional[list]=None,
                     offset: int=0,
                     limit: Optional[int]=None,
                     component_type: Optional[str]=None):
        self.portal_id      = portal_id
        self.project_id     = project_id
        self.component_type = None

        logs = await super().filter(term=term, columns=columns, offset=offset, limit=limit)  # noqa: E501

        if component_type:
            logs = [l for l in logs if l['component_type'] == component_type]

        return logs

    def _set_component_info(self, record: dict):
        if 'bug_id' in record and 'task_id' in record:
            raise AssertionError('A log entry can only be a general, bug or '
                                 'task. You appear to have passed both a bug '
                                 'and task id')
        elif 'bug_id' in record:
            self.component_id = record['bug_id']
            self.component    = 'bug'
        elif 'task_id' in record:
            self.component_id = record['task_id']
            self.component    = 'task'
        else:
            self.component    = 'general'

    async def insert(self, record: dict):
        self.portal_id  = record.pop('portal_id')
        self.project_id = record.pop('project_id')
        self._set_component_info(record)
        return await super().insert(record=record)

    async def update(self, record: dict):
        self.portal_id  = record.pop('portal_id')
        self.project_id = record.pop('project_id')
        self._set_component_info(record)
        return await super().update(record=record)

    async def delete(self,
                     id: Union[int, str],
                     *,
                     portal_id: Union[str, int],
                     project_id: Union[str, int],
                     bug_id: Optional[str]=None,
                     task_id: Optional[str]=None):
        self.portal_id  = portal_id
        self.project_id = project_id
        self._set_component_info({'bug_id': bug_id, 'task_id': task_id})
        return await super().delete(id=id)



from typing import Union, Optional

from talkzoho.projects.base_resource import BaseResource


class ProjectResource(BaseResource):

    def module_url(self, module_name):
        return '{base_url}/portal/{portal_id}/projects/{project_id}/{module}/'.format(  # noqa
            base_url=self.service.base_url,
            portal_id=self.portal_id,
            project_id=self.project_id,
            module=module_name)

    async def get(self,
                  id: Union[int, str], *,
                  portal_id: Union[str, int],
                  project_id: Union[str, int],
                  columns=None):
        self.portal_id = portal_id
        self.project_id = project_id
        return await super().get(id=id, columns=columns)

    async def filter(self, *,
                     portal_id: Union[str, int],
                     project_id: Union[str, int],
                     term: Optional[str]=None,
                     columns: Optional[list]=None,
                     offset: int=0,
                     limit: Optional[int]=None):
        self.portal_id  = portal_id
        self.project_id = project_id
        return await super().filter(
            term=term,
            columns=columns,
            offset=offset,
            limit=limit)

    async def insert(self, record: dict):
        self.portal_id  = record.pop('portal_id')
        self.project_id = record.pop('project_id')
        return await super().insert(record=record)

    async def update(self, record: dict):
        self.portal_id  = record.pop('portal_id')
        self.project_id = record.pop('project_id')
        return await super().update(record=record)

    async def delete(self,
                     id: Union[int, str],
                     *,
                     portal_id: Union[str, int],
                     project_id: Union[str, int]):
        self.portal_id  = portal_id
        self.project_id = project_id
        return await super().delete(id=id)



import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


from .books.books_client import BooksClient  # noqa
from .crm.crm_client import CRMClient  # noqa
from .projects.projects_client import ProjectsClient  # noqa



from math import ceil

from typing import Optional, Union

from urllib.parse import urlencode

from tornado.web import HTTPError
from tornado.httpclient import HTTPError as HTTPClientError
from tornado.escape import json_decode

from fuzzywuzzy import fuzz

from talkzoho import logger
from talkzoho.resource import Resource


class BaseResource(Resource):

    @property
    def module_url(self):
        return '{base_url}/{module}'.format(
            base_url=self.service.base_url,
            module=self.name)

    @property
    def base_query(self):
        return self.service.base_query

    async def get(self,
                  id: Union[int, str],
                  *,
                  organization_id: Union[str, int],
                  columns=None):
        url = '{base_url}/{id}?{query}'.format(
            base_url=self.module_url,
            id=id,
            query=urlencode({
                'organization_id': organization_id,
                **self.base_query}))

        try:
            logger.info('GET: {}'.format(url))
            response  = await self.http_client.fetch(url, method='GET')
        except HTTPClientError as http_error:
            http_code = http_error.code
            response  = json_decode(http_error.response.body.decode("utf-8"))
            message   = str(response['code']) + ': ' + response['message']
            raise HTTPError(http_code, reason=message)
        else:
            response  = json_decode(response.body.decode("utf-8"))
            results   = [v for k, v in response.items() if k not in ['code', 'message']]  # noqa

            if len(results) != 1:
                ValueError('More then one resource was returned.')

            return results[0]

    async def filter(self, *,
                     organization_id: Union[str, int],
                     term: Optional[str]=None,
                     columns: Optional[list]=None,
                     offset: int=0,
                     limit: Optional[int]=None):
        if limit == 0:
            return []
        elif not term and limit and limit <= self.service.MAX_PAGE_SIZE:
            batch_size = limit
        else:
            batch_size = self.service.MAX_PAGE_SIZE

        paging     = True
        page_index = max(ceil(offset / batch_size), 1)
        results    = []

        # Loop until we reach index we need, unless their is a search term.
        # If search term we need all records.
        while paging and (term or not limit or len(results) < limit):
            url = '{base_url}?{query}'.format(
                base_url=self.module_url,
                query=urlencode({
                    'organization_id': organization_id,
                    'per_page': batch_size,
                    'page': page_index,
                    **self.base_query}))

            try:
                logger.info('GET: {}'.format(url))
                response = await self.http_client.fetch(url, method='GET')
            except HTTPClientError as http_error:
                http_code = http_error.code
                body      = http_error.response.body
                response  = json_decode(body.decode("utf-8"))
                message   = str(response['code']) + ': ' + response['message']
                raise HTTPError(http_code, reason=message)
            else:
                response    = json_decode(response.body.decode("utf-8"))
                results    += response[self.name]
                page_index += 1
                paging      = response['page_context']['has_more_page']

        def fuzzy_score(items):
            values = [str(v).lower() for v in items.values() if v]
            target = ' '.join(values)
            return fuzz.partial_ratio(term, target)

        if term:
            results = sorted(results, key=fuzzy_score, reverse=True)

        results = results[:limit]
        if columns:
            return [{k: pl[k] for k in columns if k in columns}
                for pl in results]
        else:
            return results





class Resource:

    def __init__(self, *, service, name):
        self.service = service
        self.name    = name

    @property
    def http_client(self):
        return self.service.http_client



from talkzoho.service_client import ServiceClient
from talkzoho.books.base_resource import BaseResource


class BooksClient(ServiceClient):

    SCOPE = 'booksapi'
    MAX_PAGE_SIZE = 200

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def base_url(self):
        return 'https://books.zoho.{region}/api/v3'.format(
            region=self.region)

    @property
    def base_query(self):
        return {
            'scope': self.SCOPE,
            'authtoken': self.auth_token}

    def __getattr__(self, attr):
        """
        Translates attribute access to Zoho Projects module names.
        e.g. .price_books becomes pricebooks.
        """
        components    = attr.split('_')
        resource_name = ''.join(c.lower() for c in components)
        return BaseResource(service=self, name=resource_name)



from tornado.httpclient import AsyncHTTPClient
from talkzoho.regions import US


class ServiceClient:

    def __init__(self, *, auth_token: str, region: str=US):
        self.auth_token = auth_token
        self.region     = region

    @property
    def http_client(self):
        return AsyncHTTPClient()



from talkzoho.service_client import ServiceClient
from talkzoho.projects.base_resource import BaseResource
from talkzoho.projects.portal_resource import PortalResource
from talkzoho.projects.project_resource import ProjectResource
from talkzoho.projects.log_resource import LogResource


class ProjectsClient(ServiceClient):

    MAX_PAGE_SIZE = 100

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    @property
    def base_url(self):
        return 'https://projectsapi.zoho.{region}/restapi'.format(
            region=self.region)

    @property
    def base_query(self):
        return {'authtoken': self.auth_token}

    def __getattr__(self, attr):
        """
        Translates attribute access to Zoho Projects module names.
        e.g. . becomes Leads and custom_module_8 becomes CustomModule8
        """
        components    = attr.split('_')
        resource_name = ''.join(c.lower() for c in components)
        if resource_name in ['portals']:
            return BaseResource(service=self, name=resource_name)
        elif resource_name in ['projects', 'users']:
            return PortalResource(service=self, name=resource_name)
        elif resource_name in ['forums', 'events', 'milestones', 'tasklists', 'tasks', 'bugs']:  # noqa
            return ProjectResource(service=self, name=resource_name)
        elif resource_name in ['logs']:
            return LogResource(service=self, name=resource_name)
        else:
            raise AttributeError('Unsupported resource: ' + attr)



from typing import Optional, Union

from urllib.parse import urlencode

from tornado.web import HTTPError
from tornado.escape import json_decode

from fuzzywuzzy import fuzz

from talkzoho import logger
from talkzoho.resource import Resource
from talkzoho.crm.utils\
    import select_columns, unwrap_items, wrap_items, make_module_id_name


class CRMResource(Resource):

    def module_url(self, module_name):
        return '{base_url}/{module}'.format(
            base_url=self.service.base_url,
            module=module_name)

    @property
    def base_query(self):
        return self.service.base_query

    async def get_canonical_map(self):
        """
        Will return the module map associated to
        the Module's instance name.
        Zoho's canonical names will take precidence and user
        aliases second.
        e.g.
        The Potentials module in Zoho has been renamed opportunities
        and crm.
        """
        name = self.name
        maps = await self.service.get_module_maps()
        if self.service.use_module_aliases:
            [m] = [m for m in maps if m.plural_alias.replace(' ', '') == name]
        else:
            [m] = [m for m in maps if m.canonical_name == name]
        return m

    async def get(self, id: Union[int, str], *, columns=None):
        module_map  = await self.get_canonical_map()
        module_name = module_map.canonical_name
        module_url  = self.module_url(module_name)

        query = {
            'id': id,
            'version': 2,
            'newFormat': 2,
            **self.base_query}

        if columns:
            query['selectColumns'] = select_columns(module_name, columns)

        url = '{module_url}/getRecordById?{query}'.format(
            module_url=module_url,
            query=urlencode(query))

        logger.info('GET: {}'.format(url))
        response = await self.http_client.fetch(url, method='GET')
        body     = json_decode(response.body.decode('utf-8'))

        [item] = unwrap_items(body)
        return item

    async def insert(self, record: dict, *, trigger_workflows: bool=True):
        module_map  = await self.get_canonical_map()
        module_name = module_map.canonical_name
        module_url  = self.module_url(module_name)
        xml_record = wrap_items(record, module_name=module_name)

        url  = '{module_url}/insertRecords'.format(module_url=module_url)
        body = urlencode({
            'id': id,
            'version': 2,
            'xmlData': xml_record,
            'newFormat': 2,
            'wfTrigger': str(trigger_workflows).lower(),
            'duplicateCheck': 1,
            **self.base_query})

        logger.info('POST: {}, BODY: {}'.format(url, body))
        response = await self.http_client.fetch(url, method='POST', body=body)
        body     = json_decode(response.body.decode('utf-8'))

        [item] = unwrap_items(body)
        return item['Id']

    async def filter(self, *,
                     term: Optional[str]=None,
                     columns: Optional[list]=None,
                     offset: int=0,
                     limit: Optional[int]=None):
        module_map  = await self.get_canonical_map()
        module_name = module_map.canonical_name
        module_url  = self.module_url(module_name)
        url = '{module_url}/getRecords'.format(module_url=module_url)

        if limit == 0:
            return []
        elif not term and limit and limit <= self.service.MAX_PAGE_SIZE:
            batch_size = limit
        else:
            batch_size = self.service.MAX_PAGE_SIZE

        paging     = True
        from_index = offset + 1  # Zoho indexes at one not zero
        to_index   = offset + batch_size
        results    = []

        # Loop until we reach index we need, unless their is a search term.
        # If search term we need all records.
        while paging and (term or limit is None or to_index <= limit):
            query = {
                'fromIndex': from_index,
                'toIndex': to_index,
                'newFormat': 2,
                'version': 2,
                **self.base_query}

            if columns:
                query['selectColumns'] = select_columns(module_name, columns)

            url = '{module_url}/getRecords?{query}'.format(
                module_url=module_url,
                query=urlencode(query))

            logger.info('GET: {}'.format(url))
            response = await self.http_client.fetch(url, method='GET')
            body     = json_decode(response.body.decode('utf-8'))

            try:
                items = unwrap_items(body)
            except HTTPError as http_error:
                # if paging and hit end suppress error
                # unless first request caused the 404
                if http_error.status_code == 404 and from_index - 1 != offset:
                    paging = False
                else:
                    raise
            else:
                results   += items
                from_index = to_index + 1
                to_index  += batch_size

        def fuzzy_score(resource):
            values = [str(v) for v in resource.values() if v]
            target = ' '.join(values)
            return fuzz.partial_ratio(term, target)

        if term:
            results = sorted(results, key=fuzzy_score, reverse=True)

        return results[:limit]

    async def update(self,
                     record: dict, *,
                     trigger_workflow: bool=True):
        module_map  = await self.get_canonical_map()
        module_name = module_map.canonical_name
        module_url  = self.module_url(module_name)
        module_key  = make_module_id_name(module_map=module_map)
        record_id  = record.pop(module_key)
        xml_record = wrap_items(record, module_name=module_name)

        url  = '{module_url}/updateRecords'.format(module_url=module_url)
        body = urlencode({
            'version': 2,
            'newFormat': 2,
            'duplicateCheck': 1,
            'wfTrigger': str(trigger_workflow).lower(),
            'id': record_id,
            'xmlData': xml_record,
            **self.base_query})

        logger.info('POST: {}, BODY: {}'.format(url, body))
        response = await self.http_client.fetch(url, method='POST', body=body)
        body     = json_decode(response.body.decode('utf-8'))

        [item] = unwrap_items(body)
        return item['Id']

    async def delete(self, id: Union[int, str]):
        module_map  = await self.get_canonical_map()
        module_name = module_map.canonical_name
        module_url  = self.module_url(module_name)

        query = {'id': id, **self.base_query}
        url   = '{module_url}/deleteRecords?{query}'.format(
            module_url=module_url,
            query=urlencode(query))

        logger.info('DELETE: {}'.format(url))
        response = await self.http_client.fetch(url, method='GET')
        body     = json_decode(response.body.decode('utf-8'))

        success = unwrap_items(body)
        return success

    async def upload_file(self, *, record_id: str, url: str):
        module_map  = await self.get_canonical_map()
        module_name = module_map.canonical_name
        module_url  = self.module_url(module_name)

        url  = '{module_url}/uploadFile'.format(module_url=module_url)
        body = urlencode({'id': id, 'attachmentUrl': url, **self.base_query})

        logger.info('POST: {}, BODY: {}'.format(url, body))
        response = await self.http_client.fetch(url, method='POST', body=body)
        body     = json_decode(response.body.decode('utf-8'))

        [item] = unwrap_items(body)
        return item['Id']

    async def delete_file(self, id: Union[int, str]):
        module_map  = await self.get_canonical_map()
        module_name = module_map.canonical_name
        module_url  = self.module_url(module_name)

        query = {'id': id, **self.base_query}
        url   = '{module_url}/deleteFile?{query}'.format(
            module_url=module_url,
            query=urlencode(query))

        logger.info('GET: {}'.format(url))
        response = await self.http_client.fetch(url, method='GET')
        body     = json_decode(response.body.decode('utf-8'))

        [item] = unwrap_items(body)
        return item



from setuptools import setup
from setuptools import find_packages

try:
    from pypandoc import convert_file
    long_description = convert_file('README.md', 'rst')
except ImportError:
    long_description = open('README.md').read()


setup(
    name='talkzoho',
    version='2.0.1',
    description='Asynchronous wrapper for Zoho\'s numerous APIs',
    long_description=long_description,
    url='https://github.com/A2Z-Cloud/Talk-Zoho',
    packages=find_packages(exclude=('tests', 'tests.*')),
    author='James Stidard',
    author_email='james.stidard@a2zcloud.com',
    keywords=['talkzoho', 'Zoho', 'async', 'tornado'],
    install_requires=[
        'fuzzywuzzy',
        'python-Levenshtein',
        'tornado'])



from typing import Optional, Union

from urllib.parse import urlencode

from tornado.web import HTTPError
from tornado.httpclient import HTTPError as HTTPClientError
from tornado.escape import json_decode

from fuzzywuzzy import fuzz

from talkzoho import logger
from talkzoho.resource import Resource
from talkzoho.projects.utils import unwrap_items, to_zoho_value


class BaseResource(Resource):

    def module_url(self, module_name):
        return '{base_url}/{module}/'.format(
            base_url=self.service.base_url,
            module=module_name)

    @property
    def base_query(self):
        return self.service.base_query

    async def get(self, id: Union[int, str], *, columns=None):
        url = '{module_url}{id}/?{query}'.format(
            module_url=self.module_url(self.name),
            id=id,
            query=urlencode(self.base_query))

        logger.info('GET: {}'.format(url))
        response = await self.http_client.fetch(url, method='GET')
        body     = json_decode(response.body.decode("utf-8"))

        [item] = unwrap_items(body)
        return {k: v for k, v in item.items() if not columns or k in columns}

    async def filter(self, *,
                     term: Optional[str]=None,
                     columns: Optional[list]=None,
                     offset: int=0,
                     limit: Optional[int]=None):
        if limit == 0:
            return []
        elif not term and limit and limit <= self.service.MAX_PAGE_SIZE:
            batch_size = limit
        else:
            batch_size = self.service.MAX_PAGE_SIZE

        paging     = True
        from_index = offset + 1  # Zoho indexes at one not zero
        to_index   = offset + batch_size
        results    = []

        while paging and (term or limit is None or to_index <= limit):
            query = {
                'index': from_index,
                'range': batch_size,
                **self.base_query}

            url = '{module_url}?{query}'.format(
                module_url=self.module_url(self.name),
                query=urlencode(query))

            logger.info('GET: {}'.format(url))
            response = await self.http_client.fetch(url, method='GET')
            body     = json_decode(response.body.decode('utf-8'))

            try:
                items = unwrap_items(body)
            except HTTPError as http_error:
                # if paging and hit end suppress error
                # unless first request caused the 204
                if http_error.status_code == 204 and from_index - 1 != offset:
                    paging = False
                else:
                    raise
            else:
                results   += items
                from_index = to_index + 1
                to_index  += batch_size

        def fuzzy_score(resource):
            values = [str(v) for v in resource.values() if v]
            target = ' '.join(values)
            return fuzz.partial_ratio(term, target)

        if term:
            results = sorted(results, key=fuzzy_score, reverse=True)

        return results[:limit]

    async def insert(self, record: dict):
        url    = self.module_url(self.name)
        record = {k: to_zoho_value(v) for k, v in record.items()}
        body   = urlencode({**record, **self.base_query})

        logger.info('POST: {}, BODY: {}'.format(url, body))
        response = await self.http_client.fetch(url, method='POST', body=body)
        body     = json_decode(response.body.decode('utf-8'))

        [item] = unwrap_items(body)
        return item['id']

    async def update(self, record: dict):
        url  = '{module_url}{id}/'.format(
            module_url=self.module_url(self.name),
            id=record.pop('id'))
        body = urlencode({**record, **self.base_query})

        logger.info('POST: {}, BODY: {}'.format(url, body))
        response = await self.http_client.fetch(url, method='POST', body=body)
        body     = json_decode(response.body.decode('utf-8'))

        [item] = unwrap_items(body)
        return item

    async def delete(self, id: Union[int, str]):
        url = '{module_url}{id}/?{query}'.format(
            module_url=self.module_url(self.name),
            id=id,
            query=urlencode(self.base_query))

        logger.info('DELETE: {}'.format(url))
        # mimic zoho crm response and always return True
        # TODO: add get test to zoho CRM
        try:
            await self.http_client.fetch(url, method='DELETE')
        except HTTPClientError:
            pass
        return True
